#Updated code to sniff BLE packets from the SPHERE wearable.
#Author: Dallan Byrne 
#Date: 17/02/2016

#IMPORT Libraries
import fcntl
import os
import struct
import argparse
import subprocess
import sys
import time
import re
from datetime import datetime
from ctypes import (CDLL, get_errno)
from ctypes.util import find_library
from socket import (
    socket,
    inet_ntoa,
    SOCK_DGRAM,
    AF_INET,
    AF_BLUETOOTH,
    SOCK_RAW,
    BTPROTO_HCI,
    SOL_HCI,
    HCI_FILTER,
)




#Sniffer class
class packetSniff:
    #Class variables- Set to class defaults.
  
    timeStart = 0
    verboseFlag = False
   
    stopEarlyFlag = False
    startTime = None
    currentTime = None
    sockConn = None
  
    #Constructor
    def __init__(self):

        print('Raspberry Pi ID: ' + self.getMac())

    

    
    #---------------------------------------------------------------------------------------------------------------------------------#
    #                            Get PI Details
    #---------------------------------------------------------------------------------------------------------------------------------#
    #Get system time.        
    def getTimeUTC(self):
        return(datetime.utcnow().strftime('%Y-%m-%dT%H:%M:%S.%fZ')[:-3])
    
    def getTimeInt(self):
        return(int(round(time.time() * 1000)))
    
    # Extract Pi Mac address
    def getMac(self):
        interface = 'eth0'
        # Return the MAC address of interface
        try:
            str = open('/sys/class/net/%s/address' %interface).readline()
        except:
            str = "00:00:00:00:00:00"
        return str[0:17]  
    
    def get_ip_address(self):
        interface = 'eth0'
        s = socket(AF_INET, SOCK_DGRAM)
        return inet_ntoa(fcntl.ioctl(
            s.fileno(),
            0x8915,  # SIOCGIFADDR
            struct.pack('256s', interface[:15])
        )[20:24])  
        
        
    #---------------------------------------------------------------------------------------------------------------------------------#
    #                          String, int and conversion type manipulation
    #---------------------------------------------------------------------------------------------------------------------------------#    
    # Remove whitespaces at start and end of string - Remember strings are immutable in Python, can't pass by reference    
    def removeWhiteSpaces(self, stringWithLeadingTrailingSpaces):
        stringWithLeadingTrailingSpaces = stringWithLeadingTrailingSpaces.lstrip()                    #Remove whitespace at start
        if(~(stringWithLeadingTrailingSpaces.endswith(' '))):                        #Ensure there is a whitespace at end
            stringWithLeadingTrailingSpaces = stringWithLeadingTrailingSpaces+''
        return(stringWithLeadingTrailingSpaces)
    
    #Decode the accelerometer data from the wearable
    def decodeAccelerometerData(self, stringHexA, stringHexB):
        ''''Convert the bytes to integer and compute the accelerometer readings.
        Proprietry code. See manufacturer of wearable for details to decode 
         your specific acc data from a packet'''
        
    #Convert hex string to int
    def convertHexToInt(self, stringHex):
        return(int('0x' + stringHex, 0))    
    
    def bytes2Str(self, byteArray):
        return(''.join("{0:02x}".format(x) for x in byteArray))
    
    
    def seperateBytes(self,byteStr, seperatorStr):
        return(seperatorStr.join(byteStr[i:i+2] for i in range(0, len(byteStr), 2)))
        


    #---------------------------------------------------------------------------------------------------------------------------------#
    #                           Raspberry Pi Unix commands
    #---------------------------------------------------------------------------------------------------------------------------------#
    #Restart the Pi.
    def restartPi(self):
        command = "/usr/bin/sudo /sbin/shutdown -r now"
        self.executeCmd(command)
        
    #Reset the USB connections and the bluetooth libraries
    def bleReset(self, bluetoothDriverStr):
        #Gets the devfs path to a BLE Dongle by scraping the output
        #of the lsusb command
        
        #The lsusb command outputs a list of USB devices attached to a computer
        #in the format:
        #   Bus 001 Device 004: ID 0a5c:21e8 Broadcom Corp.
        #The devfs path to these devices is:
        #    /dev/bus/usb/<busnum>/<devnum>
        #So for the above device, it would be:
        #    /dev/bus/usb/002/009
        #Then reset the usb with the str above
        
        
        # Equivalent of the _IO('U', 20) constant in the linux kernel.
        USBDEVFS_RESET = ord('U') << (4*2) | 20
       
        proc = subprocess.Popen(['lsusb'], stdout=subprocess.PIPE)
        out = proc.communicate()[0]
        lines = out.split('\n')
        for line in lines:
            if bluetoothDriverStr in line:
                parts = line.split()
                bus = parts[1]
                dev = parts[3][:3]
        
        strCmd = ('/dev/bus/usb/%s/%s' % (bus, dev))
        
       
        #Sends the USBDEVFS_RESET IOCTL to a USB device.
        fd = os.open(strCmd, os.O_WRONLY)
        try:
            fcntl.ioctl(fd, USBDEVFS_RESET, 0)
        except:
            returnValue = -1
            print("Cannot restart the USB hub/port - "+ strCmd)
            postID = self.collectionObj.insert_one(self.getPost(self.get_ip_address(),'USB RESET ISSUE - ERROR REBOOT RPI', 100, [0,0],[0,0],[0,0],[0,0],[0,0],[0,0]))
            #Restart the pi
            self.closeDBConn()
            self.restartPi()
        else:
            returnValue = 1
            print("Restarting - "+ strCmd)       
        finally:
            os.close(fd)
            
        self.executeCmd('sleep 0.2')
        #Reset the bluetooth driver
        self.executeCmdAlt('sudo  /etc/init.d/bluetooth  restart')
        self.executeCmd('sleep 0.3')
        #Reset the BLE Dongle.     
        self.executeCmd('sudo hciconfig hci0 down');
        self.executeCmd('sudo hciconfig hci0 up');
        
        
        return(returnValue)
    
    # Execute a bash command and pipe the output to print or another function.         
    #cmdStr = the command, funcCmd is executed only if it is defined.
    def executeCmd(self,cmdStr):
        print('Shell Cmd: '+cmdStr)
        output = subprocess.check_call(cmdStr, shell=True)
        if(output!=0):
            print(output)
    #Use for certain restart commands
    def executeCmdAlt(self,cmdStr):
        print('Shell Cmd: '+cmdStr)
        p = subprocess.Popen(cmdStr, stdout=subprocess.PIPE, stderr = None, shell=True)
        retcode = p.wait()
        output = p.communicate()[0]
        p.stdout.close()

        if(output!=0):
            print(output)
           
    #---------------------------------------------------------------------------------------------------------------------------------#
    #                         BLE Sniffing commands
    #---------------------------------------------------------------------------------------------------------------------------------#
    
        
    def receiveParse(self):
       
        sphereInd = 20
        macStartInd = 6;
        macEndInd = 12;
        
        data = bytearray(self.sockConn.recv(1024))
      
        dataLen = len(data)
        if(dataLen> sphereInd):
  
            sphereStr = self.bytes2Str(data[sphereInd:sphereInd+1])
            #''.join("{0:02x}".format(x) for x in data[sphereInd:sphereInd+1])
            #SPHERE packet?
            if(sphereStr== 'f0'):
                accA_X = []
                accA_Y = []
                accA_Z = []
            
                accB_X = []
                accB_Y = []
                accB_Z = []
                
                #Start a thread for each
                # print bluetooth address from LE Advert. packet
                macStr =self.seperateBytes(self.bytes2Str(data[macEndInd:macStartInd:-1]),':');
                
                #':'.join("{0:02x}".format(x) for x in data[12:6:-1])
                
            
                rssiStr = self.bytes2Str(data[dataLen-1:dataLen])
                #''.join("{0:02x}".format(x) for x in data[dataLen-1:dataLen])
                rssiDB = (int('0x' + rssiStr, 0)-255)
                
                
                
                #This is wearable dependent so is not included here as a key security system was used.
                accPacketStr = self.seperateBytes(self.bytes2Str(data[sphereInd+1:dataLen-1]),' ');
                accPacketByteStrList = re.sub("[^\w]", " ",  accPacketStr).split()
                
               
                #Get the acc data.
                #===============================================================
                # accA_X.append(self.decodeAccelerometerData(accPacketByteStrList[0],accPacketByteStrList[1]))
                # accA_Y.append(self.decodeAccelerometerData(accPacketByteStrList[2],accPacketByteStrList[3]))
                # accA_Z.append(self.decodeAccelerometerData(accPacketByteStrList[4],accPacketByteStrList[5]))
                # accB_X.append(self.decodeAccelerometerData(accPacketByteStrList[6],accPacketByteStrList[7]))
                # accB_Y.append(self.decodeAccelerometerData(accPacketByteStrList[8],accPacketByteStrList[9]))
                # accB_Z.append(self.decodeAccelerometerData(accPacketByteStrList[10],accPacketByteStrList[11]))
                # accA_X.append(self.decodeAccelerometerData(accPacketByteStrList[12],accPacketByteStrList[13]))
                # accA_Y.append(self.decodeAccelerometerData(accPacketByteStrList[14],accPacketByteStrList[15]))
                # accA_Z.append(self.decodeAccelerometerData(accPacketByteStrList[16],accPacketByteStrList[17]))
                # accB_X.append(self.decodeAccelerometerData(accPacketByteStrList[18],accPacketByteStrList[19]))
                # accB_Y.append(self.decodeAccelerometerData(accPacketByteStrList[20],accPacketByteStrList[21]))
                # accB_Z.append(self.decodeAccelerometerData(accPacketByteStrList[22],accPacketByteStrList[23]))   
                #===============================================================
                
                
             
                timeStamp = self.getTimeUTC()
                #cntr = cntr+1;
                if(self.verboseFlag):
                    print(timeStamp + '::'+ str(macStr) +' - ' +  str(sphereStr)  +' - ' +str(rssiDB))
            
                return True    
         
         
        #If we get here then there is a missed packet 
        #print('MISSED PACKET => '+ self.bytes2Str(data))
        return False
            
            
    def setupBlueZSocket(self):
        if not os.geteuid() == 0:
            sys.exit("script only works as root")
        
        btlib = find_library("bluetooth")
        if not btlib:
            raise Exception(
                "Can't find required bluetooth libraries"
                " (need to install bluez)"
            )
        
        #Reset the BLE Dongle.     
        self.executeCmd('sudo hciconfig hci0 down');
        self.executeCmd('sudo hciconfig hci0 up');
        
        
        bluez = CDLL(btlib, use_errno=True)
        dev_id = bluez.hci_get_route(None)
        
        sock = socket(AF_BLUETOOTH, SOCK_RAW, BTPROTO_HCI)
        sock.bind((dev_id,))
        
        err = bluez.hci_le_set_scan_parameters(sock.fileno(), 0, 0x10, 0x10, 0, 0, 1000);
        #if err < 0:
        #raise Exception("Set scan parameters failed")
            # occurs when scanning is still enabled from previous call
        
        # allows LE advertising events
        hci_filter = struct.pack(
            "<IQH", 
            0x00000010, 
            0x4000000000000000, 
            0
        )
        sock.setsockopt(SOL_HCI, HCI_FILTER, hci_filter)
        
        err = bluez.hci_le_set_scan_enable(
            sock.fileno(),
            1,  # 1 - turn on;  0 - turn off
            0, # 0-filtering disabled, 1-filter out duplicates
            1000  # timeout
        )  
        self.sockConn = sock     

    #Main sniffer code.
    def startSniffer(self):
        self.startTime = sniff.getTimeInt()
        self.setupBlueZSocket()
        #Sniff
        print('Begin sniffing') 
        
        breakDownCntr = 0
        while(True):
            self.currentTime = self.getTimeInt()
            
            
            try:
                sniff.receiveParse()
            except:
                breakDownCntr+=1
                print('Error - restart the BLE driver')
                postID = self.collectionObj.insert_one(self.getPost(self.get_ip_address(),'BLE BREAKDOWN', 100, [0,0],[0,0],[0,0],[0,0],[0,0],[0,0]))     

                self.sockConn.close()
                if(breakDownCntr>3 and  (self.currentTime - self.startTime) < 30*1000):
                    print('Error - PULL THE PLUG!! RESTART THE PI')
                    postID = self.collectionObj.insert_one(self.getPost(self.get_ip_address(),'ERROR Reboot PI', 100, [0,0],[0,0],[0,0],[0,0],[0,0],[0,0]))     

                    self.restartPi()
                     
                self.bleReset('Broadcom')
                self.setupBlueZSocket()
                print('Restart sniffing') 
                
    
    
    
        
if __name__ == '__main__':
    
    sniff = packetSniff()
    
    #-------------------------------------------------------------------------------------------------#
    #                                    PARSE INPUT ARGUMENTS
    #-------------------------------------------------------------------------------------------------#
    #Parse arguments
    parser = argparse.ArgumentParser(
        description='BLE sniffer for Sphere Devices - Dallan Byrne 2016')
    # Add arguments
    parser.add_argument(
        '-v', '--verbose', help='List the sniffed packets.', action='store_true', required=False)
    parser.add_argument(
        '-t', '--testExc', help='Test the program, run for 100 packets.', action='store_true', 
        required=False)
    
    # Array for all arguments passed to script
    args = parser.parse_args()
    # Assign to class variables
    sniff.verboseFlag = args.verbose
    sniff.stopEarlyFlag = args.testExc

    print('Begin Packet Dumping ')
    print('======================================================')
    if( sniff.verboseFlag):
      
        print('\t Listing Packets as instructed...')
    else:
        print('\t Running in Silent Mode...')
    print('\t\t\t--------')    
    if( sniff.stopEarlyFlag):
        print('\t Run only 100 packets - test mode...')
    else:
        print('\t Running forever...')
    
    print('======================================================')
    
    #-------------------------------------------------------------------------------------------------#
    #                                    Start Sniffing
    #-------------------------------------------------------------------------------------------------#
    sniff.startSniffer()

 
    
    
    
        